#include <sta/config.hpp>
#ifdef STA_CAN_BUS_ENABLE

#include <sta/tacos/can_bus.hpp>
#include <sta/debug/assert.hpp>
// TODO fix this shit
extern CAN_HandleTypeDef hcan1;
namespace sta
{
	namespace tacos
	{
		CanBus::CanBus()
			: TacosThread{"Can Bus", STA_TACOS_CAN_BUS_PRIORITY},
			  canBusController_(new STM32CanController(&STA_STM32_CAN_HANDLE)),
			  canBusSysQueue_(STA_RTOS_CAN_BUS_QUEUE_LENGTH),
			  canBusDataQueue_(STA_RTOS_CAN_BUS_QUEUE_LENGTH),
              canBus_{canBusController_, HAL_GetTick, sta::tacos::handleSysMessage, sta::tacos::handleDataMessage}
		{
		}

        void CanBus::init()
	    {
		    canBusController_->start();
	    }

        void CanBus::func()
        {
        	messageEvent.clear(STA_RTOS_CAN_ANY);
            uint32_t flags = messageEvent.wait(STA_RTOS_CAN_ANY, osWaitForever);

            STA_DEBUG_PRINTLN("CanBus received Flag");
            if (flags != static_cast<uint32_t>(osErrorTimeout))
            {
                STA_ASSERT_MSG((flags & osStatusReserved) == flags, "Unexpected error occurred in wait");

                if (flags & STA_RTOS_CAN_FLAG_SYS_QUEUED)
                {
                    // Take messages from queue until empty
                    CanSysMsg msg;
                    while (CanBus::_instance->getCanBusMsg(&msg, 0))
                    {
                    	STA_DEBUG_PRINTLN("CanBus about to send");
                        canBus_.send(msg);
                    }
                }

                if (flags & STA_RTOS_CAN_FLAG_DATA_QUEUED)
                {
                    // Take messages from queue until empty
                    CanDataMsg msg;
                    while (CanBus::_instance->getCanBusMsg(&msg, 0))
                    {
                        canBus_.send(msg);
                    }

					canBus_.tx_.processFrame();
                }

                if (flags & STA_RTOS_CAN_FLAG_MSG_AVAIL)
                {
                    STA_DEBUG_PRINTLN("[event] CAN INT");

                    canBus_.processRx();
                }

                if (flags & STA_RTOS_CAN_FLAG_SHOW_STATS)
                {
                    canBus_.showStatistics();
                }
            }

			// Synchronus message check
			canBus_.processRx();

            // Process ISOTP transmissions
            canBus_.processTx();
        }

        bool CanBus::queueCanBusMsg(const CanDataMsg& msg, uint32_t timeout)
		{
			STA_ASSERT((msg.header.sid & STA_CAN_SID_SYS_BITS) == 0);
			STA_ASSERT(msg.header.payloadLength <= sizeof(msg.payload));

			if (canBusDataQueue_.put(msg, timeout))
			{
				// Signal task
				messageEvent.set(STA_RTOS_CAN_FLAG_DATA_QUEUED);
				messageEvent.clear(STA_RTOS_CAN_ANY);
				return true;
			}
			else
			{
				return false;
			}
		}

		bool CanBus::queueCanBusMsg(const CanSysMsg& msg, uint32_t timeout)
		{
			STA_ASSERT((msg.header.sid & ~STA_CAN_SID_SYS_BITS) == 0);

			if (canBusSysQueue_.put(msg, timeout))
			{
				// Signal tasek
				messageEvent.set(STA_RTOS_CAN_FLAG_SYS_QUEUED);
				messageEvent.clear(STA_RTOS_CAN_ANY);
				return true;
			}
			else
			{
				return false;
			}
		}


		bool CanBus::getCanBusMsg(CanDataMsg * msg, uint32_t timeout)
		{
			return (canBusDataQueue_.get(msg, timeout));
		}

		bool CanBus::getCanBusMsg(CanSysMsg * msg, uint32_t timeout)
		{
			return (canBusSysQueue_.get(msg, timeout));
		}

		CanBus* CanBus::_instance = nullptr;

		RtosEvent CanBus::messageEvent;

    } /* namespace tacos */
} /* namespace sta */

namespace sta {
	namespace tacos
	{
		void handleSysMessage(const sta::CanRxHeader & header, const uint8_t * payload)
		{
			// Write frame payload to DebugSerial
			STA_DEBUG_PRINTLN("[event] RX sys frame");

			debug::printFrameID(header.id);
			debug::printPayloadHex(payload, header.payloadLength);

			// Sysmessage is mainly only state change from GRSM
			// TODO add other cases

			// 0 is from state, 1 is to state, 2 is lockout, 3 is failsafe (-1 if inactive)
			if(payload[1] > 0 && payload[1] < STA_TACOS_NUM_STATES){

				if(payload[3] == -1){
					sta::tacos::Statemachine::instance()->requestStateTransition(payload[0], payload[1], payload[2]);
				}
				else{
					sta::tacos::Statemachine::instance()->requestTimedStateTransition(payload[0], payload[1], payload[2], payload[3]);
				}
			}
		}

		void handleDataMessage(const sta::IsotpMessage & msg)
		{
			STA_ASSERT(msg.buffer != nullptr);
			STA_ASSERT(msg.size != 0);

			STA_DEBUG_PRINTLN("[event] RX data message");

			debug::printFrameID(msg.frameID);

			// TODO Forward message to other threads

	//		if (buffer[0] == DEMO_BMP_PACKET_ID)
	//		{
	//			BmpPacket packet;
	//			if (unpack(buffer + 1, size - 1, &packet))
	//			{
	//				STA_DEBUG_PRINTLN();
	//				STA_DEBUG_PRINTLN("# ############");
	//				STA_DEBUG_PRINTLN("# ## BMP380 ##");
	//				STA_DEBUG_PRINTLN("# ############");
	//				STA_DEBUG_PRINTLN("#");
	//
	//				STA_DEBUG_PRINT("# temperature: ");
	//				STA_DEBUG_PRINT(packet.temperature);
	//				STA_DEBUG_PRINTLN(" *C");
	//				STA_DEBUG_PRINT("# pressure: ");
	//				STA_DEBUG_PRINT(packet.pressure);
	//				STA_DEBUG_PRINTLN(" Pa");
	//				STA_DEBUG_PRINT("# altitude: ");
	//				STA_DEBUG_PRINT(packet.altitude);
	//				STA_DEBUG_PRINTLN(" m");
	//				STA_DEBUG_PRINTLN();
	//			}
	//			else
	//			{
	//				STA_DEBUG_PRINTLN("[error] BMP unpack failed");
	//			}
	//		}
	//		else
			{
				STA_DEBUG_PRINT("ID: ");
				STA_DEBUG_PRINTLN(msg.buffer[0], sta::IntegerBase::HEX);
				STA_DEBUG_PRINT("size: ");
				STA_DEBUG_PRINTLN(msg.size);
			}
		}
	} // namespace tacos
} // namespace sta

#endif // STA_CAN_BUS_ENABLE
